#include <torch/extension.h>
#include <ATen/cuda/CUDAContext.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <cuda_runtime.h>

namespace py = pybind11;

// This kernel fuses the elementwise addition of the identity and the main branch, and applies ReLU
// It evenly distributes the workload across threads and blocks using a grid-stride loop
__global__ void fused_add_relu_kernel(const float* __restrict__ identity, float* __restrict__ out, int size) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int stride = blockDim.x * gridDim.x;
    for (; idx < size; idx += stride) {
        float val = out[idx] + identity[idx];
        out[idx] = (val > 0.0f) ? val : 0.0f;
    }
}

// The bottleneck function implements the main branch and optionally a downsample branch.
// The final residual addition and activation are fused via a custom CUDA kernel to evenly distribute work.
torch::Tensor bottleneck_fn(
    torch::Tensor x,
    const torch::Tensor& conv1_w,
    const torch::Tensor& conv2_w,
    const torch::Tensor& conv3_w,
    const torch::Tensor& bn1_w,
    const torch::Tensor& bn1_b,
    const torch::Tensor& bn1_m,
    const torch::Tensor& bn1_v,
    const torch::Tensor& bn2_w,
    const torch::Tensor& bn2_b,
    const torch::Tensor& bn2_m,
    const torch::Tensor& bn2_v,
    const torch::Tensor& bn3_w,
    const torch::Tensor& bn3_b,
    const torch::Tensor& bn3_m,
    const torch::Tensor& bn3_v,
    const torch::Tensor& downsample_conv_w,
    const torch::Tensor& downsample_bn_w,
    const torch::Tensor& downsample_bn_b,
    const torch::Tensor& downsample_bn_m,
    const torch::Tensor& downsample_bn_v,
    int64_t stride,
    bool is_training
) {
    torch::Tensor identity = x;

    bool has_downsample = downsample_conv_w.defined();
    torch::Tensor downsample_out;
    if (has_downsample) {
        downsample_out = torch::conv2d(x, downsample_conv_w, /*bias=*/torch::Tensor(), stride)
            .to(x.dtype(), /*non_blocking=*/true, /*copy=*/false, torch::MemoryFormat::Contiguous);
        downsample_out = torch::batch_norm(
            downsample_out,
            downsample_bn_w,
            downsample_bn_b,
            downsample_bn_m,
            downsample_bn_v,
            is_training,
            /*momentum=*/0.1,
            /*eps=*/1e-5,
            /*cudnn_enabled=*/true
        );
        identity = downsample_out;
    }

    torch::Tensor out = torch::conv2d(x, conv1_w, /*bias=*/torch::Tensor())
        .to(x.dtype(), /*non_blocking=*/true, /*copy=*/false, torch::MemoryFormat::Contiguous);
    out = torch::batch_norm(
        out,
        bn1_w,
        bn1_b,
        bn1_m,
        bn1_v,
        is_training,
        0.1,
        1e-5,
        true
    );
    out = torch::relu(out);

    out = torch::conv2d(out, conv2_w, /*bias=*/torch::Tensor(), stride, /*padding=*/1)
        .to(x.dtype(), /*non_blocking=*/true, /*copy=*/false, torch::MemoryFormat::Contiguous);
    out = torch::batch_norm(
        out,
        bn2_w,
        bn2_b,
        bn2_m,
        bn2_v,
        is_training,
        0.1,
        1e-5,
        true
    );
    out = torch::relu(out);

    out = torch::conv2d(out, conv3_w, /*bias=*/torch::Tensor())
        .to(x.dtype(), /*non_blocking=*/true, /*copy=*/false, torch::MemoryFormat::Contiguous);
    out = torch::batch_norm(
        out,
        bn3_w,
        bn3_b,
        bn3_m,
        bn3_v,
        is_training,
        0.1,
        1e-5,
        true
    );

    // Fuse the residual addition and ReLU activation using the custom CUDA kernel
    int64_t num_elements = out.numel();
    const int threads = 256;
    const int blocks = (num_elements + threads - 1) / threads;
    fused_add_relu_kernel<<<blocks, threads, 0, at::cuda::getCurrentCUDAStream()>>>(
        identity.data_ptr<float>(),
        out.data_ptr<float>(),
        num_elements
    );
    cudaDeviceSynchronize(); // Ensure kernel execution completes before proceeding

    return out;
}

// The forward function processes the ResNet101 architecture layer by layer.
// It prefetches all parameters on the correct device and uses the bottleneck_fn to apply each block.
// The fused kernel ensures that the residual addition and activation are distributed evenly among CUDA threads.
torch::Tensor forward(
    torch::Tensor x,
    py::object params,
    bool is_training
) {
    auto device = x.device();

    // Initial layer parameters
    auto conv1_w = params.attr("get")(py::str("conv1_w")).cast<torch::Tensor>().contiguous().to(device, true);
    auto bn1_w = params.attr("get")(py::str("bn1_w")).cast<torch::Tensor>().contiguous().to(device, true);
    auto bn1_b = params.attr("get")(py::str("bn1_b")).cast<torch::Tensor>().contiguous().to(device, true);
    auto bn1_m = params.attr("get")(py::str("bn1_m")).cast<torch::Tensor>().contiguous().to(device, true);
    auto bn1_v = params.attr("get")(py::str("bn1_v")).cast<torch::Tensor>().contiguous().to(device, true);

    x = torch::conv2d(x, conv1_w, /*bias=*/torch::Tensor(), /*stride=*/2, /*padding=*/3)
        .to(x.dtype(), /*non_blocking=*/true, /*copy=*/false, torch::MemoryFormat::Contiguous);
    x = torch::batch_norm(
        x,
        bn1_w,
        bn1_b,
        bn1_m,
        bn1_v,
        is_training,
        0.1,
        1e-5,
        true
    );
    x = torch::relu(x);
    x = torch::max_pool2d(x, /*kernel_size=*/3, /*stride=*/2, /*padding=*/1);

    // Process each layer's blocks
    for (int layer_idx = 1; layer_idx <= 4; ++layer_idx) {
        std::string layer_key = "layer" + std::to_string(layer_idx) + "_blocks";
        py::list blocks = params.attr("get")(py::str(layer_key)).cast<py::list>();
        for (size_t block_idx = 0; block_idx < blocks.size(); ++block_idx) {
            py::object block_params = blocks[block_idx];

            auto conv1_w = block_params.attr("get")(py::str("conv1_w")).cast<torch::Tensor>().contiguous().to(device, true);
            auto conv2_w = block_params.attr("get")(py::str("conv2_w")).cast<torch::Tensor>().contiguous().to(device, true);
            auto conv3_w = block_params.attr("get")(py::str("conv3_w")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn1_w = block_params.attr("get")(py::str("bn1_w")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn1_b = block_params.attr("get")(py::str("bn1_b")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn1_m = block_params.attr("get")(py::str("bn1_m")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn1_v = block_params.attr("get")(py::str("bn1_v")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn2_w = block_params.attr("get")(py::str("bn2_w")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn2_b = block_params.attr("get")(py::str("bn2_b")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn2_m = block_params.attr("get")(py::str("bn2_m")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn2_v = block_params.attr("get")(py::str("bn2_v")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn3_w = block_params.attr("get")(py::str("bn3_w")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn3_b = block_params.attr("get")(py::str("bn3_b")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn3_m = block_params.attr("get")(py::str("bn3_m")).cast<torch::Tensor>().contiguous().to(device, true);
            auto bn3_v = block_params.attr("get")(py::str("bn3_v")).cast<torch::Tensor>().contiguous().to(device, true);

            torch::Tensor downsample_conv_w = torch::Tensor();
            torch::Tensor downsample_bn_w = torch::Tensor();
            torch::Tensor downsample_bn_b = torch::Tensor();
            torch::Tensor downsample_bn_m = torch::Tensor();
            torch::Tensor downsample_bn_v = torch::Tensor();
            
            if (py::bool_(block_params.attr("__contains__")(py::str("downsample_conv_w")))) {
                downsample_conv_w = block_params.attr("get")(py::str("downsample_conv_w")).cast<torch::Tensor>().contiguous().to(device, true);
                downsample_bn_w = block_params.attr("get")(py::str("downsample_bn_w")).cast<torch::Tensor>().contiguous().to(device, true);
                downsample_bn_b = block_params.attr("get")(py::str("downsample_bn_b")).cast<torch::Tensor>().contiguous().to(device, true);
                downsample_bn_m = block_params.attr("get")(py::str("downsample_bn_m")).cast<torch::Tensor>().contiguous().to(device, true);
                downsample_bn_v = block_params.attr("get")(py::str("downsample_bn_v")).cast<torch::Tensor>().contiguous().to(device, true);
            }

            int64_t stride = (block_idx == 0 && layer_idx > 1) ? 2 : 1;
            x = bottleneck_fn(
                x,
                conv1_w,
                conv2_w,
                conv3_w,
                bn1_w,
                bn1_b,
                bn1_m,
                bn1_v,
                bn2_w,
                bn2_b,
                bn2_m,
                bn2_v,
                bn3_w,
                bn3_b,
                bn3_m,
                bn3_v,
                downsample_conv_w,
                downsample_bn_w,
                downsample_bn_b,
                downsample_bn_m,
                downsample_bn_v,
                stride,
                is_training
            );
        }
    }

    x = torch::adaptive_avg_pool2d(x, {1, 1}).contiguous();
    x = x.view({x.size(0), -1});

    auto fc_w = params.attr("get")(py::str("fc_w")).cast<torch::Tensor>().contiguous().to(device, true);
    auto fc_b = params.attr("get")(py::str("fc_b")).cast<torch::Tensor>().contiguous().to(device, true);
    return torch::linear(x, fc_w, fc_b);
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "ResNet101 forward function with fused addition and ReLU");
}
